import time
from functools import wraps

import numpy as np
from scipy.special import gammaln

ComplexArr = np.typing.NDArray[np.complexfloating]


def corr(a, b):
    """Cross-correlation coefficient."""
    return np.corrcoef(a.flat, b.flat)[0, 1]


def complex_corr(truth: ComplexArr, recon_image: ComplexArr):
    """Accepts truth and data that are 2D complex-valued reconstructions"""

    # for convenience
    term = lambda img: np.sqrt(np.sum(np.square(np.abs(img))))

    # define num and denom
    num = np.sum(truth.conj() * recon_image)
    denom = term(truth) * term(recon_image)

    return np.abs(num / denom)


def rmse(a, b):
    """root-mean square error"""
    return np.sqrt(np.mean(np.square(a - b)))


def safelog(x, x_min=1e-30):
    """avoids division by zero for log(0)"""
    return np.log(np.clip(x, x_min, None))


def log_factorials(n):
    """log factorials for log of 0, 1!, ..., upto n!"""
    return gammaln(np.arange(1, n + 1))


def timeit(method):
    """custom wrapper for estimating time required for a function to execute"""

    @wraps(method)
    def wrapper(*args, **kwargs):
        start_time = time.time()
        result = method(*args, **kwargs)
        end_time = time.time()
        print(f"{method.__name__} => {(end_time-start_time):.3f} sec")

        return result

    return wrapper


def radial_average(image, bins=100, max_radius=None, return_radius=False):
    """Radial average for a given image upto to
    radius = image.shape // 2

    Args:
        image (np.ndarray): 2D image
        bins (int, optional): number of bins to calculate radial average. Defaults to 100.
        return_radius (bool, optional): _description_. Defaults to False.

    Returns:
        _type_: _description_
    """

    # check if square image for simplicity
    M, N = image.shape
    assert M == N

    if max_radius is None:
        max_radius = N / 2

    y, x = np.indices(image.shape)
    distances = np.sqrt((x - max_radius) ** 2 + (y - max_radius) ** 2)
    hist, edges = np.histogram(np.ravel(distances), bins=bins, weights=np.ravel(image))

    r = 0.5 * (edges[:-1] + edges[1:])

    if return_radius:
        return r, hist

    return hist


def psd(image, bins, ft_switch=False):
    """computing Power spectral density of an fft image

    Args:
        image (np.ndarray): Either a normal or fft image
        bins (int): no. of bins to averge radially
        ft_switch (bool): image should be fft, if not compute it.
    """
    if ft_switch:
        image = np.fft.fftshift(np.fft.fftn(image))

    return radial_average(np.square(np.abs(image)), bins)


def crop_image(image, zoom_factor=1.5):
    """just a function to crop an object to focus on the object"""

    # make sure new shape is smaller in every index
    img_shape = np.array(image.shape)
    new_shape = np.array(img_shape // zoom_factor, dtype=int)
    assert np.all(img_shape >= new_shape)

    dy, dx = img_shape - new_shape
    sy = slice(dy // 2, img_shape[0] - dy // 2)
    sx = slice(dx // 2, img_shape[1] - dx // 2)

    new_image = image.copy()
    return new_image[sy, sx]



# implements the EM methods: uncensored EM, Heuristic EM and full EM

import abc
from typing import TypeAlias

import numpy as np
from scipy.stats import poisson

Vector: TypeAlias = list[float]


class ConventionalMEF:
    """
    Considers the uncensored data (ignores overexposure) from the overexposed dataset
    to estimate maximum likelihood (MLE), background is subtracted from this estimate/data.

    Args:
        data (np.ndarray): 3D array with the first index for acquisition
            times i and the remaining for an image M x N. Therefore i x M x N
        threshold (float): censoring threshold
        times (list): acquisition times
        background (np.ndarray or scalar, optional): background with same shape
            as data or a scalar value. Defaults to 0.
    """

    def __init__(
        self,
        data: np.ndarray,
        threshold: float,
        times: list,
        background: np.ndarray = None,
    ) -> None:

        self.data = data
        self.threshold = threshold
        self.times = times
        self.background = background if background is not None else 0
        self.fused_image = np.ones(data.shape[1:])

    @property
    def label(self):
        """
        returns binary mask m(q) that labels
        censored counts as 0 and uncensored counts as 1
        """

        data = self.data.copy()
        mask = data < self.threshold
        return np.array(mask)

    def mle(self):
        """fused estimate based on mle estimate"""

        # subtract background and clip negative values
        nu = np.clip(self.data - self.background, 0, None)

        # fused image using uncensored data
        fused_image = np.sum(nu * self.label, axis=0)
        fused_image /= np.tensordot(self.times, self.label, axes=1) + 1e-3

        self.fused_image[...] = fused_image


class ExpectationMaximization(abc.ABC):
    """Abstract class for ExpectationMaximization with some common implemented methods

    To implement various types of EM algorithms for MEF

    Args:
        data (np.ndarray): 3D array with the first index for acquisition
            times i and the remaining for an image M x N. Therefore i x M x N
        threshold (float): censoring threshold
        times (list): acquisition times
        background (np.ndarray or scalar, optional): background with same shape
            as data or a scalar value. Defaults to 0.
        alpha (float, optional): gamma prior shape. Defaults to 1.
        beta (float, optional): gamma prior rate. Defaults to 1/threshold.
    """

    def __init__(
        self,
        data: np.ndarray,
        threshold: float,
        times: Vector,
        background: np.ndarray = None,
        update_fluxes: bool = False,
        max_flux_iteration: int = 100,
    ) -> None:
        self.data = data
        self.threshold = threshold
        self.times = times
        self.background = background if background is not None else 0
        self.update_fluxes = update_fluxes

        # prior gamma shape and rate
        self.alpha = 1e-3
        self.beta = 1e-3

        # max flux iterations
        self.max_flux_iterations = max_flux_iteration
        self.current_flux_iteration = 0

        if self.update_fluxes:
            self.alpha_flux = np.full_like(self.times, 1.0)
            self.beta_flux = 1 / np.array(self.times)

        # using setter/getter to set a rate (default: `prior_rate`)
        self.fused_image = None
        self.initialize()

    def initialize(self):
        """Initializes the prior rate based on gamma prior"""

        # prior results to the threshold with no data
        prior_rate = self.alpha / self.beta

        # assuming first argument corresponds to acquisition times
        prior_rate = np.full(self.data.shape[1:], prior_rate, dtype=self.data.dtype)

        self.fused_image = prior_rate

    @property
    def label(self):
        """
        returns binary mask m(q) that labels
        censored counts as 0 and uncensored counts as 1
        """

        data = self.data.copy()
        mask = data < self.threshold
        return np.array(mask)

    @property
    def _rates(self):
        """rates are scaled with times and the current fused image in the iteration"""
        rates = np.multiply.outer(self.times, self.fused_image)

        return rates

    def flux_estimate(self, nu):
        """Updates fluxes based on the current estimate of the fused image"""
        fluxes = self.alpha_flux + np.sum(nu, axis=tuple(range(1, nu.ndim)))
        fluxes /= self.beta_flux + np.sum(self.fused_image)

        return fluxes

    @abc.abstractmethod
    def e_step(self):
        """involves estimating the missing data or incorrect data"""
        pass

    @abc.abstractmethod
    def m_step(self, nu):
        """estimating the fused image with new data"""
        pass

    def __next__(self):
        """Single iteration of EM is implemented as an iterator"""

        nu = self.e_step()
        self.m_step(nu)

        # estimate fluxes if true
        if self.update_fluxes:
            # update fluxes only upto `max_flux_iterations`
            while self.current_flux_iteration < self.max_flux_iterations:
                self.times = self.flux_estimate(nu)
                self.current_flux_iteration += 1

                # exit the loop after every iteration
                break

    def run(self, n_iter: int = 200, return_params=False):
        """iteratively recover the fused image"""

        fused = [self.fused_image.copy()]
        times = [np.array(self.times.copy())]
        for _ in range(n_iter):
            next(self)
            if return_params:
                fused.append(self.fused_image.copy())
                if self.update_fluxes:
                    times.append(self.times.copy())

        # return the parameters if needed
        if return_params:
            if self.update_fluxes:
                return fused, times
            else:
                return fused


class UncensoredEM(ExpectationMaximization):
    """Uncensored EM uses only uncensored data, however, it iteratively removes
    background from the data

    Args:
        data (np.ndarray): 3D array with the first index for acquisition
            times i and the remaining for an image M x N. Therefore i x M x N
        threshold (float): censoring threshold
        times (list): acquisition times
        background (np.ndarray or scalar, optional): background with same shape
            as data or a scalar value. Defaults to 0.
        alpha (float, optional): gamma prior shape. Defaults to 1.
        beta (float, optional): gamma prior rate. Defaults to 1/threshold.
    """

    def e_step(self):
        """
        Estimating the e-step that returns the new data
        with background treatment that accepts the fused image
        """

        # background factor
        bk_factor = self._rates / (self._rates + self.background + 1e-5)

        # returns uncensored data with bk factor
        nu = bk_factor * self.label * self.data

        return nu

    def m_step(self, nu):
        """considers the new updated data with background to estimate the rate"""

        # calculate uncensored mean along exposure time axis
        fused_image = np.sum(nu, axis=0) + self.alpha
        fused_image /= np.tensordot(self.times, self.label, axes=1) + self.beta

        # copy to global image
        self.fused_image[...] = fused_image


class BayesianMEF(UncensoredEM):
    """EM uses censored and uncensored data for fusion, moreover considers the
    background treatment aswell

    Args:
        data (np.ndarray): 3D array with the first index for acquisition
            times i and the remaining for an image M x N. Therefore i x M x N
        threshold (float): censoring threshold
        times (list): acquisition times or flux factors
        background (np.ndarray or scalar, optional): background with same shape
            as data or a scalar value. Defaults to 0.
        alpha (float, optional): gamma prior shape. Defaults to 1.
        beta (float, optional): gamma prior rate. Defaults to 1/threshold
        n_init (int, optional): iterations for the initial rate (initialization)
    """

    def __init__(
        self,
        data: np.ndarray,
        threshold: float,
        times: Vector,
        background: np.ndarray = None,
        update_fluxes: bool = False,
        n_iter_fluxes: int = 100,
        n_init: int = 100,
    ) -> None:
        self.n_init = n_init

        # prior gamma shape and rate
        self.alpha = 1e-3
        self.beta = 1e-3

        # super constructor to access remaining attributes
        super().__init__(
            data, threshold, times, background, update_fluxes, n_iter_fluxes
        )

        if self.update_fluxes:
            self.alpha_flux = np.full_like(self.times, 1.0)
            self.beta_flux = 1 / np.array(self.times)

        # counter
        self.current_flux_iteration = 0

        # initializes fused image
        self.fused_image = None
        self.initialize()

    def initialize(self):
        """
        Initialize fused image by running EM iterations using
        uncensored EM
        """
        # initializes `self.fused_image` to the gamma prior
        super().initialize()

        # uncensored EM as initialization for improving
        # speed of convergence of full EM
        for _ in range(self.n_init):
            nu = super().e_step()
            super().m_step(nu)

    @staticmethod
    def censored_poison(threshold, rate):
        """
        expectation under the censored poisson model that
        estimates missing data
        """

        # truncated poisson
        poi = poisson(rate)
        num = poi.sf(threshold - 1)
        denom = poi.sf(threshold)

        # expectation of missing data

        # quick fix: returns rate when num and denom are zero
        # also avoid zero division by replacing denom zero as 1,
        # although the case where num is non-zero and denom zero should
        # never happen.
        nonzero_denom = np.where(denom == 0, 1, denom)
        cens_mean = np.where(
            (num == 0) & (denom == 0), rate, rate * num / nonzero_denom
        )

        # catching nans of censored mean
        if np.any(np.isnan(cens_mean)):
            print(
                f"Warning: NaN for censored mean at" f"{np.where(np.isnan(cens_mean))}"
            )

        return cens_mean

    def e_step(self):
        """gives an expectation for missing censored poisson data"""

        # rates per acquisition
        rates_bg = self._rates + self.background + 1e-5
        threshold = self.threshold

        # check for censored pixels per acquisition
        nu = self.data.copy()
        for i in range(len(self.times)):
            # do nothing for uncensored acquisitions (label=True)
            if np.all(self.label[i]):
                continue

            # current rate with background for censored pixels
            censored = ~self.label[i]
            rate = rates_bg[i, censored]

            # expectation under the censored poisson model for missing data
            nu[i, censored] = self.censored_poison(threshold, rate)

        # scale data with background factor
        nu *= self._rates / rates_bg

        return nu

    def m_step(self, nu):
        """The maximization step will estimate the mean with all the data"""

        # MAP estimate with the new data
        fused_image = np.sum(nu, axis=0) + self.alpha
        fused_image /= np.sum(self.times) + self.beta

        # update fused image
        self.fused_image[...] = fused_image


class Prior(abc.ABC):
    @abc.abstractmethod
    def __call__(self):
        pass


class CensoringPrior(Prior):
    """
    A gamma prior whose expected value is an image with censored pixels
    set to the threshold

    Parameters
    ----------
    data (np.ndarray): 3D array with the first index for acquisition
            times i and the remaining for an image M x N. Therefore i x M x N
    threshold : float
        censoring threshold
    """

    def __init__(self, data: np.ndarray, threshold: float) -> None:
        self.data = data
        self.threshold = threshold
        self.alpha = 1e-3
        self.beta = 1e-3

    def cens_label(self) -> np.ndarray:
        """
        returns binary mask m(q) that labels
        censored counts as 1 and uncensored counts as 0
        """

        data = self.data.copy()
        mask = data >= self.threshold
        return np.array(mask)

    def __call__(self) -> np.ndarray:
        # check the presence of censoring (nonzero sum)
        sum_label = np.sum(self.cens_label(), axis=0)

        # labels censored counts as 1 and uncensored as 0
        mask = sum_label > 0

        # shape and rate of gamma distribution
        self.alpha = self.alpha * ~mask + mask
        self.beta = self.beta * ~mask + mask / self.threshold


class PtychogramMean(CensoringPrior):
    def __init__(self, data: np.ndarray, threshold: float) -> None:
        """Exponential Gamma Prior which is a mean of the diffraction
        patterns recorded at different exposure times

        Parameters
        ----------
        data (np.ndarray): 3D array with the first index for acquisition
            times i and the remaining for an image M x N. Therefore i x M x N
        threshold : float
        censoring threshold
        """
        super().__init__(data, threshold)

    def __call__(self) -> np.ndarray:
        self.beta = self.alpha / np.mean(self.data, axis=0)



# Uses a parallel implementation for Bayesian MEF
import os
from functools import partial

import h5py
import numpy as np
from joblib import Parallel, delayed
from tqdm import tqdm

from bayes_mef.algos import BayesianMEF, ConventionalMEF, ExpectationMaximization, Prior
from bayes_mef.utils import corr, timeit


@timeit
def parallel_fusion(
    em_method: ExpectationMaximization,
    pty_stack: np.ndarray,
    background: np.ndarray,
    threshold: float,
    times: list[float],
    n_iter: int,
    cpus: int,
    prior: Prior,
    update_fluxes: bool,
    max_flux_iterations: int,
):
    """running EM algorithm parallely by allocating CPU cores"""

    # assumes the dimensions to be in the order of (scans, exposure, M, M)
    assert pty_stack.ndim == 4

    # run em partial by proving other params
    run_em_partial = partial(
        run_em,
        em_method=em_method,
        threshold=threshold,
        times=times,
        n_iter=n_iter,
        prior=prior,
        update_fluxes=update_fluxes,
        max_flux_iterations=max_flux_iterations,
    )

    print(f"\n Starting {em_method.__name__} ..")

    # stack as list of tuples of exposures of a scan with background
    pty_stack = [pattern for pattern in pty_stack]

    # assumes the same dimension as data `pty_stack` (scans, exposure, M, M)
    if background.ndim == 4:

        # creates a bg stack
        bg_stack = [bg_per_scan for bg_per_scan in background]

        with Parallel(n_jobs=cpus) as parallel:
            fused = parallel(
                delayed(run_em_partial)(patterns_per_scan, bg_per_scan)
                for patterns_per_scan, bg_per_scan in tqdm(
                    zip(pty_stack, bg_stack), total=len(pty_stack)
                )
            )

    # assumes dimension to be (exposures, M, M)
    if background.ndim == 3:
        # creates another partial function that loads same background for every scan
        run_em_partial_same_bg = partial(run_em_partial, bg=background)

        with Parallel(n_jobs=cpus) as parallel:
            fused = parallel(
                delayed(run_em_partial_same_bg)(pattern) for pattern in tqdm(pty_stack)
            )

    # unpack fused images and updated fluxes
    fused_em, times_em = zip(
        *[(fused_per_scan[0], fused_per_scan[1]) for fused_per_scan in fused]
    )

    return np.array(fused_em), np.array(times_em)


def run_em(
    stack,
    bg,
    em_method,
    threshold,
    times,
    n_iter,
    prior,
    update_fluxes,
    max_flux_iterations,
):
    """runs EM on exposures for a given scan"""

    assert stack.ndim == 3
    assert bg.ndim == 3

    # EM launcher
    em = em_method(stack, threshold, times, bg, update_fluxes, max_flux_iterations)

    # if prior exists, extract the values of alpha and beta
    # or fall back to default values.
    if prior is not None:
        assert issubclass(prior, Prior)
        em_prior = prior(stack, threshold)
        em_prior()

        # set the prior values
        em.alpha = em_prior.alpha
        em.beta = em_prior.beta

    em.run(n_iter)
    fused_em = em.fused_image.copy()
    times_em = em.times.copy()

    return fused_em, times_em


def set_fluxes(stack, threshold, use_uncensored=True):
    """
    Creating fluxes on the fly when exposure times is None,
    especially useful with experimental data due to its variations.
    """
    # stack should have (n_exposures, n_scans, X, Y) or (n_exposures, X, Y)

    if use_uncensored:
        uncens_mask = stack < threshold
        uncens_stack = stack * uncens_mask
        pty = uncens_stack.copy()
    else:
        pty = stack.copy()

    # estimate fluxes
    ind = len(pty) // 2 - 1
    axis_last = tuple(range(1, pty.ndim))
    fluxes = (np.sum(pty, axis=axis_last) + 1e-5) / (np.sum(pty[ind]) + 1e-5)

    return list(fluxes)


def set_threshold(stack):
    """sets threshold based on the exposures per scan"""

    # stack is (n_exposures, X, Y)
    return stack.max() - np.sqrt(stack.max())


class LaunchMEF:
    """Interface for launching MEF"""

    def __init__(
        self,
        pty_stack: np.ndarray,
        background: np.ndarray,
        times: list[float] = None,
        threshold: float = None,
        prior: Prior = None,
        pty_noisefree_stack: np.ndarray = None,
        update_fluxes: bool = False,
    ) -> None:

        # initialize params
        self.pty_stack = pty_stack  # (n_exposures, n_scans, X, Y)
        self.background = background
        self.threshold = (
            set_threshold(self.pty_stack) if threshold is None else threshold
        )
        self.times = (
            set_fluxes(self.pty_stack, self.threshold) if times is None else times
        )
        self.prior = prior
        self.pty_noisefree_stack = pty_noisefree_stack
        self.update_fluxes = update_fluxes

        # (n_exposures, n_scans, X, Y)
        assert self.pty_stack.ndim == 4

        # move second dimension to first (n_scans, n_exposures, X, Y)
        self.pty_stack = np.moveaxis(self.pty_stack, 1, 0)

        if self.background.ndim == 4:
            self.background = np.moveaxis(self.background, 1, 0)

    def run_mle(self):
        """runs MLE"""

        if self.background.ndim == 3:
            pty_mle = []
            for patterns in self.pty_stack:
                # Conventional MEF (MLE)
                conv_mle = ConventionalMEF(
                    patterns, self.threshold, self.times, self.background
                )
                conv_mle.mle()
                pty_mle.append(conv_mle.fused_image)

        if self.background.ndim == 4:
            pty_mle = []
            for patterns, bg in zip(self.pty_stack, self.background):
                # Conventional MEF (MLE)
                conv_mle = ConventionalMEF(patterns, self.threshold, self.times, bg)
                conv_mle.mle()
                pty_mle.append(conv_mle.fused_image)

        if self.pty_noisefree_stack is not None:
            # compare with some ground truth (if exists)
            mle_corr = corr(pty_mle[0], self.pty_noisefree_stack[0])
            print(f"Correlation of MLE with noisefree: {mle_corr:.3f}")

        # min and max for the fused data
        print(f"Min MLE: {np.min(pty_mle):.2f}," f" Max MLE: {np.max(pty_mle):.2f}")

        return np.array(pty_mle)

    def run_em(self, n_iter=200, ncpus=6):
        """Runs EM in parallel"""

        # run in parallel
        pty_em, em_fluxes = parallel_fusion(
            BayesianMEF,
            self.pty_stack,
            self.background,
            self.threshold,
            self.times,
            n_iter=n_iter,
            cpus=ncpus,
            prior=self.prior,
            update_fluxes=self.update_fluxes,
            max_flux_iterations=n_iter,
        )

        # running EM again in parallel without updating fluxes and instead using a fixed
        # average of the corrected fluxes
        if self.update_fluxes:
            corrected_mean_fluxes = list(np.mean(em_fluxes, axis=0))
            print("Corrected fluxes: ", corrected_mean_fluxes)

            pty_em, _ = parallel_fusion(
                BayesianMEF,
                self.pty_stack,
                self.background,
                self.threshold,
                corrected_mean_fluxes,
                n_iter=n_iter,
                cpus=ncpus,
                prior=self.prior,
                update_fluxes=False,
                max_flux_iterations=n_iter,
            )

        if self.pty_noisefree_stack is not None:
            # compare with some ground truth (if exists)
            em_corr = corr(pty_em[0], self.pty_noisefree_stack[0])
            print(f"Correlation of EM with noisefree: {em_corr:.3f}")

        # min and max for the fused data
        print(f"Min EM: {pty_em.min():.2f}," f" Max EM: {pty_em.max():.2f}")

        return pty_em, em_fluxes

    def save_result(
        self,
        filepath: str,
        pty_fused: np.ndarray,
        fluxes_em: np.ndarray,
        params: dict,
        pty_key: str,
        save_misc: bool,
    ):
        """saves result as hdf5"""

        if "orientation" not in params:
            params.update({"orientation": 0})

        with h5py.File(filepath, "w") as hf:
            hf.create_dataset(pty_key, data=pty_fused, dtype="f")
            hf.create_dataset("exposures", data=self.times, dtype="f")
            hf.create_dataset("fluxes", data=fluxes_em)

            if save_misc:
                hf.create_dataset("ptychogram_noisy", data=self.pty_stack, dtype="f")
                hf.create_dataset("threshold", data=self.threshold, dtype="f")
                hf.create_dataset("background", data=self.background, dtype="f")

            if self.pty_noisefree_stack is not None:
                hf.create_dataset(
                    "ptychogram_noNoise", data=self.pty_noisefree_stack, dtype="f"
                )

            hf.create_dataset("encoder", data=params["encoder"], dtype="f")
            hf.create_dataset("binningFactor", data=params["binningFactor"], dtype="i")
            hf.create_dataset("dxd", data=(params["dxd"],), dtype="f")
            hf.create_dataset("Nd", data=(params["Nd"],), dtype="i")
            # hf.create_dataset("No", data=(params["No"],), dtype="i")
            hf.create_dataset("zo", data=(params["zo"],), dtype="f")
            hf.create_dataset("wavelength", data=(params["wavelength"],), dtype="f")
            hf.create_dataset(
                "entrancePupilDiameter",
                data=(params["entrancePupilDiameter"],),
                dtype="f",
            )

            hf.create_dataset("orientation", data=params["orientation"])

        print(f"Saved fused result at {filepath}")

    def run_and_save(
        self,
        savepath: str,
        filename: str,
        params: dict,
        pty_key: str = "ptychogram",
        mle_switch: bool = True,
        em_switch: bool = True,
        n_iter_em: int = 200,
        ncpus_em: int = 6,
        save_misc: bool = False,
        separate_fused_dir: bool = False,
    ):

        # runs and saves MLE result
        if mle_switch:
            if separate_fused_dir:
                filepath_mle = f"{savepath}/mle_fused/mle_{filename}.hdf5"
                os.makedirs(os.path.dirname(filepath_mle), exist_ok=True)
            else:
                filepath_mle = f"{savepath}/mle_{filename}.hdf5"

            if not os.path.exists(filepath_mle):
                pty_mle = self.run_mle()
                mle_fluxes = self.times
                self.save_result(
                    filepath_mle,
                    pty_mle,
                    mle_fluxes,
                    params,
                    pty_key,
                    save_misc,
                )
            else:
                print(f"File {filepath_mle} already exists")

        # runs and saves EM result
        if em_switch:
            if separate_fused_dir:
                filepath_em = f"{savepath}/em_fused/em_{filename}.hdf5"
                os.makedirs(os.path.dirname(filepath_em), exist_ok=True)
            else:
                filepath_em = f"{savepath}/em_{filename}.hdf5"

            if not os.path.exists(filepath_em):
                pty_em, fluxes_em = self.run_em(n_iter_em, ncpus_em)
                self.save_result(
                    filepath_em,
                    pty_em,
                    fluxes_em,
                    params,
                    pty_key,
                    save_misc,
                )
            else:
                print(f"File {filepath_em} already exists")
